### =========================================================================
### GWASdata objects
### -------------------------------------------------------------------------

### Slots:
###
###   o datapath, dataconn: the path and connection to a NetCDF file
###     containing the GWAS experimental data (a matrix). See the man page
###     for GWASdata for what the NetCDF file should contain.
###     Note that the primary keys of the 'subjects' and 'snps' tables
###     contained in the SQLite db file described below must be indices into
###     the rows and cols of this GWAS matrix, respectively.
###
###   o metadatapath, metadataconn: the path to an SQLite db file containing
###     the GWAS metadata. See the man page for GWASdata for what the SQLite
###     db should contain.
###
###   o nrow, ncol: the dimensions of the GWAS matrix.
###
setClass("GWASdata",
    representation(
        ## The GWAS experimental data:
        datapath="character",
        dataconn="list",  # "ncdf" cannot be used here
        ## The GWAS metadata:
        metadatapath="character",
        metadataconn="SQLiteConnection",
        ## The dimensions of the data:
        nrow="integer",
        ncol="integer"
    )
)


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Constructor.
###

GWASdata <- function (datapath, metadatapath) 
{
    dataconn <- open.ncdf(datapath, write = FALSE)
    ## Need to drop the class attribute ("ncdf", an S3 class), otherwise
    ## new() will complain that the "dataconn" slot is not a list.
    class(dataconn) <- NULL
    metadataconn <- dbConnect(SQLite(), dbname = metadatapath)
    nrow <- dbGetQuery(metadataconn, "SELECT count(*) FROM subjects")[[1]]
    ncol <- dbGetQuery(metadataconn, "SELECT count(*) FROM snps")[[1]]
    new("GWASdata", datapath = datapath, dataconn = dataconn, 
        metadatapath = metadatapath, metadataconn = metadataconn,
        nrow = nrow,  ncol = ncol)
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Accessors.
###
### Only getters, no setters (all the GWASdata slots must be considered
### read-only slots).
###

setGeneric("datapath", function(x) standardGeneric("datapath"))

setMethod("datapath", "GWASdata", function(x) x@datapath)

setGeneric("dataconn", function(x) standardGeneric("dataconn"))

setMethod("dataconn", "GWASdata",
    function(x)
    {
        dataconn <- x@dataconn
        ## Restore the class attribute that was dropped by the constructor.
        class(dataconn) <- "ncdf"
        dataconn
    }
)

setGeneric("metadatapath", function(x) standardGeneric("metadatapath"))

setMethod("metadatapath", "GWASdata", function(x) x@metadatapath)

setGeneric("metadataconn", function(x) standardGeneric("metadataconn"))

setMethod("metadataconn", "GWASdata", function(x) x@metadataconn)

### No setGeneric statement for nrow, ncol and dim. Those functions belong
### to base R and the setMethod statements below will implicitely turn
### them into generics when the StudentGWAS package is loaded.

setMethod("nrow", "GWASdata", function(x) x@nrow)

setMethod("ncol", "GWASdata", function(x) x@ncol)

setMethod("dim", "GWASdata", function(x) c(nrow(x), ncol(x)))

setGeneric("getCols", signature = "x",
    function(x, first, last = first) standardGeneric("getCols")
)

setMethod("getCols", "GWASdata",
    function(x, first, last = first)
        getGWAScols(dataconn(x), first, last)
)


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Validity.
###
### Note that a validity method should NOT raise an error.
###

.valid.GWASdata.dataconn <- function(object)
{
    dataconn <- dataconn(object)
    ## See R/utils.R for the details about the getNcdfVarSummary()
    ## function.
    nc_vs <- getNcdfVarSummary(dataconn)
    if (!("snpData" %in% nc_vs$name))
        return("variable \"snpData\" not found in NetCDF file")
    ## There might be other variables in addition to the "snpData" var.
    nc_vs <- nc_vs[nc_vs$name == "snpData", ]
    if (nc_vs$type != "byte") {
        msg <- c("NetCDF variable \"snpData\" must be of type \"byte\" ",
                 "(found \"", nc_vs$type, "\")")
        return(paste(msg, collapse=""))
    }
    if (nc_vs$ndims != 2) {
        msg <- c("NetCDF variable \"snpData\" must have 2 dimensions ",
                 "(found ", nc_vs$ndims, ")")
        return(paste(msg, collapse=""))
    }
    if (nc_vs$dim1_name != "subjectDim" || nc_vs$dim2_name != "snpDim") {
        msg <- c("dimensions of NetCDF variable \"snpData\" must ",
                 "be named \"subjectDim\" and \"snpDim\" ",
                 "(found \"", nc_vs$dim1_name, "\" and \"",
                 nc_vs$dim2_name, "\")")
        return(paste(msg, collapse=""))
    }
    if (nc_vs$dim1 != nrow(object) || nc_vs$dim2 != ncol(object)) {
        msg <- c("dimensions of NetCDF variable \"snpData\" ",
                 "don't match values of slots 'nrow' and 'ncol'")
        return(paste(msg, collapse=""))
    }
    NULL
}

.valid.GWASdata.metadataconn <- function(object)
{
    metadataconn <- metadataconn(object)
    if (!all(c("subjects", "snps") %in% dbListTables(metadataconn)))
        return("table \"subjects\" or \"snps\" not found in SQLite db")

    ## Check fields in table 'subjects'.
    REQUIRED_SUBJECTS_FIELDS <- c("id", "subject_id", "case_control")
    subjects_fields <- dbListFields(metadataconn, "subjects")
    missing_fields <- setdiff(REQUIRED_SUBJECTS_FIELDS, subjects_fields)
    if (length(missing_fields) != 0) {
        msg <- c("fields not found in SQL table \"subjects\": ",
                 paste(missing_fields, collapse=", "))
        return(paste(msg, collapse=""))
    }

    ## Check fields in table 'snps'.
    REQUIRED_SNPS_FIELDS <- c("id", "snp_id", "seqnames", "start", "gene_id")
    snps_fields <- dbListFields(metadataconn, "snps")
    missing_fields <- setdiff(REQUIRED_SNPS_FIELDS, snps_fields)
    if (length(missing_fields) != 0) {
        msg <- c("fields not found in SQL table \"snps\": ",
                 paste(missing_fields, collapse=", "))
        return(paste(msg, collapse=""))
    }

    nrow <- dbGetQuery(metadataconn, "SELECT count(*) FROM subjects")[[1]]
    ncol <- dbGetQuery(metadataconn, "SELECT count(*) FROM snps")[[1]]
    if (nrow != nrow(object) || ncol != ncol(object)) {
        msg <- c("nb of rows in tables \"subjects\" and \"snps\" ",
                 "don't match values of slots 'nrow' and 'ncol'")
        return(paste(msg, collapse=""))
    }
    NULL
}

setValidity("GWASdata",
    function(object)
    {
        ## Check the 'datapath' slot.
        if (!is.character(datapath(object)) ||
            length(datapath(object)) != 1 ||
            is.na(datapath(object)))
            return("'datapath' slot must be a single string")

        ## Check the 'metadatapath' slot.
        if (!is.character(metadatapath(object)) ||
            length(metadatapath(object)) != 1 ||
            is.na(metadatapath(object)))
            return("'metadatapath' slot must be a single string")

        ## Check the 'nrow' slot.
        if (!is.integer(nrow(object)) ||
            length(nrow(object)) != 1 ||
            is.na(nrow(object)) ||
            nrow(object) < 0)
            return("'nrow' slot must be a single non-negative integer")

        ## Check the 'ncol' slot.
        if (!is.integer(ncol(object)) ||
            length(ncol(object)) != 1 ||
            is.na(ncol(object)) ||
            ncol(object) < 0)
            return("'ncol' slot must be a single non-negative integer")

        ## Check the 'dataconn' slot.
        msg <- .valid.GWASdata.dataconn(object)
        if (!is.null(msg))
            return(msg)

        ## Check the 'metadataconn' slot.
        msg <- .valid.GWASdata.metadataconn(object)
        if (!is.null(msg))
            return(msg)
        TRUE
    }
)


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Coercion.
###

setAs("GWASdata", "matrix", function(from) getCols(from, 1, ncol(from)))

setMethod("as.matrix", "GWASdata", function(x) as(x, "matrix"))


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### "show" method.
###

setMethod("show", "GWASdata",
    function(object)
    {
        nrow <- nrow(object)
        ncol <- ncol(object)
        cat(class(object), "instance with", nrow, "subjects and",
            ncol, "SNPs\n")
        cat("| datapath: ", datapath(object), "\n", sep="")
        cat("| metadatapath: ", metadatapath(object), "\n", sep="")
    }
)

