/* 

   Composite linkage disequilibrium is a statistical measure of
   association across loci, when haplotype phase is not known. See
   Weir and Cockerham, 1989; Weir 1996 pp. 94, 125.

   Delta = n - p_{A_1} p_{B_1}

   where
   
   N: number of subjects
   p_{A_1}: frequency of allele A_1 at snp A
   p_{B_1}: frequency of an allele B_1 at snp B
   n: X_{A_1A_1, B_1B_1} + 
       (1/2) * (X_{A_1A_1, B_1B_2} + X_{A_1A_2, B_1B_1}) + 
         (1/4) X_{A_1A_2, B_1B_2}

   WEIR, B., 1996 Genetic Data Analysis II. Sinauer Associates,
   Sunderland, MA.
       
   WEIR, B. S., and C. C. COCKERHAM, 1989 Complete characterization of
   disequilibrium at two loci, pp. 86–110 in Mathematical Evolutionary
   Theory, edited by M. W. FELDMAN. Princeton University Press,
   Princeton, NJ.

 */

#include "composite_linkage_disequilibrium.h"

/*
   Calculate the composite linkage disequilibrium for 2 SNPs.
   'snp1' and 'snp2' must be vectors of genotypes (one for each SNP)
   containing 1 byte per subject. The byte values are assumed to be:
     0 for missing;
     1 for homozygous major (A/A);
     2 for heterozygous (A/B);
     3 for homozygous minor (B/B).
*/
static double cld_two_snps(const unsigned char *snp1,
		const unsigned char *snp2, int nsub)
{
	double X[9]; /* 9 genotype frequencies */
	double p_A1, p_B1, n_A1_B1, N;
	int i, k;

	/* ...tally 9 genotype frequencies */
	for (k = 0; k < 9; ++k)
		X[k] = 0;
	for (i = 0; i < nsub; ++i)
		if (snp1[i] & snp2[i])
			X[ (snp1[i] - 1) * 3 + snp2[i] - 1] += 1;

	/* ...and calculate composite linkage disequilibrium */
	N = 0;
	for (k = 0; k < 9; ++k) /* total # individuals */
		N += X[k];
	for (k = 0; k < 9; ++k) /* genotype _frequencies_ */
		X[k] /= N;
	p_A1 = X[0] + X[1] + X[2] + 0.5 * (X[3] + X[4] + X[5]);
	p_B1 = X[0] + X[3] + X[6] + 0.5 * (X[1] + X[4] + X[7]);
	n_A1_B1 = X[0] + 0.5 * (X[1] + X[3]) + 0.25 * X[4];
	return n_A1_B1 - p_A1 * p_B1;
}

/* .C entry point */
void composite_linkage_disequilibrium(
		const unsigned char *snp, /* matrix subject x snp */
		const int *n_sub, /* # subjects */
		const int *n_snp, /* # snps */
		const int *width, /* adjacent snp window */
		double *delta) /* result */
{
    /* convenience and index variables */
    int nsub = n_sub[0], nsnp = n_snp[0], w = width[0];
    int j1, j2;

    /* for each combination of 2 SNPs... */
    for (j1 = 0; j1 < nsnp - w; ++j1)
	for (j2 = j1 + 1; j2 <= j1 + w; ++j2)
	    *(delta++) = cld_two_snps(snp + j1 * nsub, snp + j2 * nsub, nsub);
}

/* .Call entry point */
SEXP composite_linkage_disequilibrium_1(SEXP snp_r, SEXP width_r)
{
    /* check incoming types */
    if (!IS_INTEGER(width_r) ||
        LENGTH(width_r) != 1 ||
        INTEGER(width_r)[0] == NA_INTEGER)
    {
		error("'width' must be a single non-NA integer");
    }
    if (!IS_RAW(snp_r) || LENGTH(GET_DIM(snp_r)) != 2)
		error("'snp' must be matrix of raw()");

    /* retrieve inputs */
    int width = INTEGER(width_r)[0],
		n_sub = INTEGER(GET_DIM(snp_r))[0],
		n_snp = INTEGER(GET_DIM(snp_r))[1];

    /* allocate memory for return */
    SEXP delta_r;
    PROTECT(delta_r = allocMatrix(REALSXP, width, n_snp - width));

    /* do the calculation */
    composite_linkage_disequilibrium(RAW(snp_r), 
									 &n_sub, &n_snp, &width,
									 REAL(delta_r));

    /* unprotect and return */    
    UNPROTECT(1);
    return delta_r;
}
